#!/usr/bin/env python

# This file is part of Androguard.
#
# Copyright (C) 2010, Geoffroy Gueguen <geoffroy.gueguen@gmail.com>
# All rights reserved.
#
# Androguard is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Androguard is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Androguard.  If not, see <http://www.gnu.org/licenses/>.

import sys
sys.path.append('./')
import androguard
import analysis
import Instruction
import Structure
import Util
import copy


class This():
    def __init__(self, cls):
        self.cls = cls
        self.used = False
        self.content = self

    def get_type(self):
        return self.cls.name

    def get_value(self):
        return 'this'

    def get_name(self):
        return 'this'


class Register():
    def __init__(self, content):
        self.content = content
        self.nbuses = 0
        self.used = False
        self.isPair = False

    def modify(self, ins):
        if self.used:
            self.dump(ins)
        self.content = ins
        self.nbuses = 0
        self.used = False

    def get_content(self):
        self.nbuses += 1
        if self.nbuses > 1:
            self.used = True
            Util.log('GET CONTENT -> USED TRUE (%s)' % self.content, 'debug')
        return self.content

    def dump(self, ins):
        Util.log("""Register Dump :
                ---------------
                Old value :
                %s
                -> %s
                -------
                New value :
                %s
                -> %s (used : %s)
                ---------------""" % (str(self.content),
                self.content.get_value(), str(ins),
                ins.get_value(), self.used), 'debug')

    def __deepcopy__(self, dic=None):
        d = dic.get(self, None)
        if d is None:
            r = Register(self.content)
            r.nbuses = self.nbuses
            r.used = self.used
            r.isPair = self.isPair
            dic[self] = r
            return r
        return d

    def __str__(self):
        return 'Register :\n\tused : %s\n' \
               '\tcontent : %s\t\tvalue : %s.' % (self.used, str(self.content),
                                                  str(self.content.get_value()))

    def __repr__(self):
        return repr(self.content)


class DvMethod():
    def __init__(self, methanalysis, this):
        self.memory = {}
        self.method = methanalysis.get_method()
        self.name = self.method.get_name()
        self.lparams = []
        self.variables = Instruction.Variable()
        
        if self.name == '<init>':
            self.name = self.method.get_class_name()[1:-1].split('/')[-1]
        self.basic_blocks = methanalysis.basic_blocks.bb
        code = self.method.get_code()

        access = self.method.get_access()
        self.access = [flag for flag in Util.ACCESS_FLAGS_METHODS
                                     if flag & access]
        desc = self.method.get_descriptor()
        self.type = Util.get_type(desc.split(')')[-1])
        self.paramsType = Util.get_params_type(desc)

        Util.log('Searching loops in method %s' % self.name, 'debug')
        
        exceptions = methanalysis.exceptions.exceptions
        Util.log('METHOD : %s' % self.name, 'debug')
        #print "excepts :", [ e.exceptions for e in exceptions ]
        self.ast = Structure.ConstructAST(self.basic_blocks, exceptions)

        #androguard.bytecode.method2png('graphs/%s#%s.png' % \
        #        (self.method.get_class_name().split('/')[-1][:-1], self.name),
        #                                                         methanalysis)

        if code is None:
            Util.log('CODE NONE : %s %s' % (self.name,
                                            self.method.get_class_name()),
                                            'debug')
        else:
            start = code.registers_size.get_value() - code.ins_size.get_value()
            # 0x8 == Static : 'this' is not passed as a parameter
            if 0x8 in self.access:
                self._add_parameters(start)
            else:
                self.memory[start] = Register(this)
                self._add_parameters(start + 1)
        self.ins = []
        self.cur = 0

    def _add_parameters(self, start):
        size = 0
        for paramType in self.paramsType:
            param = self.variables.newVar(paramType)
            param.param = True
            self.memory[start + size] = Register(param)
            self.lparams.append(param)
            size += param.size

    def process(self):
        if self.ast is None:
            return
        Util.log('METHOD : %s' % self.name, 'debug')
        for child in self.ast.childs:
            Util.log('Processing %s' % child, 'debug')
            blockins = child.process(self.memory, self.variables, 0)
            if blockins:
                self.ins.append(blockins)

    def debug(self, code=None):
        if code is None:
            code = []
        Util.log('Dump of method :', 'debug')
        for j in self.memory.values():
            Util.log(j, 'debug')
        Util.log('Dump of ins :', 'debug')
        acc = []
        for i in self.access:
            if i == 0x10000:
                self.type = None
            else:
                acc.append(Util.ACCESS_FLAGS_METHODS.get(i))
        if self.type:
            proto = '    %s %s %s(' % (' '.join(acc), self.type, self.name)
        else:
            proto = '    %s %s(' % (' '.join(acc), self.name)
        if self.paramsType:
            proto += ', '.join(['%s' % param.decl() for param in self.lparams])
        proto += ')\n    {\n'
        Util.log(proto, 'debug')
        code.append(proto)
        for i in self.ins:
            Util.log('%s' % i, 'debug')
            code.append(''.join(['   ' * 2 + '%s\n' % ii for ii in
                                                            i.splitlines()]))
        Util.log('}', 'debug')
        code.append('    }')
        return ''.join(code)

    def __repr__(self):
        return 'Method %s' % self.name


class DvClass():
    def __init__(self, dvclass, bca):
        self.name = dvclass.get_name()
        self.package = dvclass.get_name().rsplit('/', 1)[0][1:].replace('/',
                                                                        '.')
        self.this = This(self)
        lmethods = [(method.get_idx(), DvMethod(bca.get_method(method),
                       self.this)) for method in dvclass.get_methods()]
        self.methods = dict(lmethods)
        self.fields = {}
        for field in dvclass.get_fields():
            self.fields[field.get_name()] = field
        self.access = []
        self.prototype = None
        self.subclasses = {}
        self.code = []
        self.inner = False

        Util.log('Class : %s' % self.name, 'log')
        Util.log('Methods added :', 'log')
        for index, meth in self.methods.iteritems():
            Util.log('%s (%s, %s)' % (index, meth.method.get_class_name(),
                                   meth.name), 'log')
        Util.log('\n', 'log')

        access = dvclass.format.get_value().access_flags
        access = [flag for flag in Util.ACCESS_FLAGS_CLASSES if flag & access]
        for i in access:
            self.access.append(Util.ACCESS_FLAGS_CLASSES.get(i))
        
        self.prototype = '%s class %s' % (' '.join(self.access), self.name)
        
        self.interfaces = dvclass._interfaces
        self.superclass = dvclass._sname

    def add_subclass(self, innername, dvclass):
        self.subclasses[innername] = dvclass
        dvclass.inner = True

    def get_methods(self):
        meths = copy.copy(self.methods)
        for cls in self.subclasses.values():
            meths.update(cls.get_methods())
        return meths

    def select_meth(self, nb):
        if nb in self.methods:
            self.methods[nb].process()
            self.code.append(self.methods[nb].debug())
        elif self.subclasses.values():
            for cls in self.subclasses.values():
                cls.select_meth(nb)
        else:
            Util.log('Method %s not found.' % nb, 'error')

    def show_code(self):
        if not self.inner and self.package:
            Util.log('package %s;\n' % self.package, 'log')

        if self.superclass is not None:
            self.superclass = self.superclass[1:-1].replace('/', '.')
            if self.superclass.split('.')[-1] == 'Object':
                self.superclass = None
            if self.superclass is not None:
                self.prototype += ' extends %s' % self.superclass
        if self.interfaces is not None:
            self.interfaces = self.interfaces[1:-1].split(' ')
            self.prototype += ' implements %s' % ', '.join(
                        [n[1:-1].replace('/', '.') for n in self.interfaces])

        Util.log('%s {' % self.prototype, 'log')
        for field in self.fields.values():
            access =  [Util.ACCESS_FLAGS_FIELDS.get(flag) for flag in
                        Util.ACCESS_FLAGS_FIELDS if flag & field.get_access()]
            type = Util.get_type(field.get_descriptor())
            name = field.get_name()
            Util.log('\t%s %s %s;' % (' '.join(access), type, name), 'log')

        Util.log('', 'log')
        for cls in self.subclasses.values():
            cls.show_code()

        for ins in self.code:
            Util.log(ins, 'log')
        Util.log('}', 'log')

    def process(self):
        Util.log('SUBCLASSES : %s' % self.subclasses, 'debug')
        for cls in self.subclasses.values():
            cls.process()
        Util.log('METHODS : %s' % self.methods, 'debug')
        for meth in self.methods:
            self.select_meth(meth)

    def __str__(self):
        return 'Class name : %s.' % self.name

    def __repr__(self):
        if self.subclasses == {}:
            return 'Class(%s)' % self.name
        return 'Class(%s) -- Subclasses(%s)' % (self.name, self.subclasses)


class DvMachine():
    def __init__(self, name):
        vm = androguard.AndroguardS(name).get_vm()
        bca = analysis.VMAnalysis(vm)
        self.vm = vm
        self.bca = bca
        ldict = [(dvclass.get_name(), DvClass(dvclass, bca))
                 for dvclass in vm.get_classes()]
        self.classes = dict(ldict)
        Util.merge_inner(self.classes)

    def get_class(self, className):
        for name, cls in self.classes.iteritems():
            if className in name:
                return cls

    def process_class(self, cls):
        if cls is None:
            Util.log('No class to process.', 'error')
        else:
            cls.process()

    def process_method(self, cls, meth):
        if cls is None:
            Util.log('No class to process.', 'error')
        else:
            cls.select_meth(meth)

    def show_code(self, cls):
        if cls is None:
            Util.log('Class not found.', 'error')
        else:
            cls.show_code()


if __name__ == '__main__':
    Util.DEBUG_LEVEL = 'debug'

    MACHINE = DvMachine('examples/android/TestsAndroguard/bin/classes.dex')

    from pprint import pprint
    temp = Util.wrap_stream()
    Util.log('===========================', 'log')
    Util.log('dvclass.get_Classes :', 'log')
    pprint(MACHINE.classes, temp)
    Util.log(temp, 'log')
    Util.log('===========================', 'log')

    CLS = raw_input('Choose a class: ')
    if CLS == '*':
        for CLS in MACHINE.classes:
            Util.log('CLS : %s' % CLS, 'log')
            cls = MACHINE.get_class(CLS)
            if cls is None:
                Util.log('%s not found.' % CLS, 'error')
            else:
                MACHINE.process_class(cls)
        Util.log('\n\nDump of code:', 'log')
        Util.log('===========================', 'log')
        for CLS in MACHINE.classes:
            MACHINE.show_code(MACHINE.get_class(CLS))
            Util.log('===========================', 'log')
    else:
        cls = MACHINE.get_class(CLS)
        if cls is None:
            Util.log('%s not found.' % CLS, 'error')
        else:
            Util.log('======================', 'log')
            temp.clean()
            pprint(cls.get_methods(), temp)
            Util.log(temp, 'log')
            Util.log('======================', 'log')
            METH = raw_input('Method: ')
            if METH == '*':
                Util.log('CLASS = %s' % cls, 'log')
                MACHINE.process_class(cls)
            else:
                MACHINE.process_method(cls, int(METH))
            Util.log('\n\nDump of code:', 'log')
            Util.log('===========================', 'log')
            MACHINE.show_code(cls)
